# SPDX-License-Identifier: Apache-2.0

from celery import Celery
from celery.signals import worker_process_init
import json
import pynetbox
from redis import Redis

from osism import settings
from osism.actions import manage_device, manage_interface
from osism.tasks import Config, openstack, run_command

app = Celery("netbox")
app.config_from_object(Config)

redis = None
nb = None


@worker_process_init.connect
def celery_init_worker(**kwargs):
    global nb
    global redis

    redis = Redis(
        host=settings.REDIS_HOST,
        port=settings.REDIS_PORT,
        db=settings.REDIS_DB,
        socket_keepalive=True,
    )
    redis.ping()

    if settings.NETBOX_URL and settings.NETBOX_TOKEN:
        nb = pynetbox.api(settings.NETBOX_URL, token=settings.NETBOX_TOKEN)

        if settings.IGNORE_SSL_ERRORS:
            import requests

            requests.packages.urllib3.disable_warnings()
            session = requests.Session()
            session.verify = False
            nb.http_session = session


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    pass


@app.task(bind=True, name="osism.tasks.netbox.periodic_synchronize_ironic")
def periodic_synchronize_ironic(self):
    """Synchronize the state of Ironic with Netbox"""
    openstack.baremetal_node_list.apply_async((), link=synchronize_device_state.s())


@app.task(bind=True, name="osism.tasks.netbox.run")
def run(self, action, arguments):
    pass


@app.task(bind=True, name="osism.tasks.netbox.update_network_interface_name")
def update_network_interface_name(self, mac_address, network_interface_name):
    manage_interface.update_network_interface_name(mac_address, network_interface_name)


@app.task(bind=True, name="osism.tasks.netbox.synchronize_device_state")
def synchronize_device_state(self, data):
    """Synchronize the state of Ironic with Netbox"""

    if type(data) == str:
        data = json.loads(data)

    if not data:
        return

    for device in data:
        manage_device.set_provision_state(device["Name"], device["Provisioning State"])
        manage_device.set_power_state(device["Name"], device["Power State"])


@app.task(bind=True, name="osism.tasks.netbox.states")
def states(self, data):
    result = manage_device.get_states(data.keys())
    return result


@app.task(bind=True, name="osism.tasks.netbox.set_state")
def set_state(self, device=None, state=None, state_type=None):
    manage_device.set_state(device, state, state_type)


@app.task(bind=True, name="osism.tasks.netbox.set_maintenance")
def set_maintenance(self, device=None, state=None):
    manage_device.set_maintenance(device, state)


@app.task(bind=True, name="osism.tasks.netbox.diff")
@app.task(bind=True, name="osism.tasks.netbox.get_devices_not_yet_registered_in_ironic")
def get_devices_not_yet_registered_in_ironic(
    self, status="active", tags=["managed-by-ironic"], ironic_enabled=True
):
    global nb

    devices = nb.dcim.devices.filter(
        tag=tags, status=status, cf_ironic_enabled=[ironic_enabled]
    )

    result = []

    for device in devices:
        if (
            "ironic_state" in device.custom_fields
            and device.custom_fields["ironic_state"] != "registered"
        ):
            result.append(device.name)

    return result


@app.task(
    bind=True,
    name="osism.tasks.netbox.get_devices_that_should_have_an_allocation_in_ironic",
)
def get_devices_that_should_have_an_allocation_in_ironic(self):
    global nb

    devices = nb.dcim.devices.filter(
        tag=["managed-by-ironic", "managed-by-osism"],
        status="active",
        cf_ironic_enabled=[True],
        cf_ironic_state=["registered"],
        cf_provision_state=["available"],
        cf_introspection_state=["introspected"],
        cf_device_type=["server"],
    )

    result = []

    for device in devices:
        result.append(device.name)

    return result


@app.task(bind=True, name="osism.tasks.netbox.manage")
def manage(self, *arguments, publish=True, locking=False, auto_release_time=3600):
    netbox_manager_env = {
        "NETBOX_MANAGER_URL": str(settings.NETBOX_URL),
        "NETBOX_MANAGER_TOKEN": str(settings.NETBOX_TOKEN),
        "NETBOX_MANAGER_IGNORE_SSL_ERRORS": str(settings.IGNORE_SSL_ERRORS),
        "NETBOX_MANAGER_VERBOSE": "true",
    }

    return run_command(
        self.request.id,
        "/usr/local/bin/netbox-manager",
        netbox_manager_env,
        *arguments,
        publish=publish,
        locking=locking,
        auto_release_time=auto_release_time
    )


@app.task(bind=True, name="osism.tasks.netbox.ping")
def ping(self):
    global nb

    status = nb.status()

    return status



# SPDX-License-Identifier: CC-BY-NC-4.0
# Copyright OSISM GmbH, 2022-2023

from loguru import logger
from pottery import Redlock

from osism import utils


def get_state(device):
    """Gets the state (device_state) stored in the Netbox for a  device."""

    result = None
    device_a = utils.nb.dcim.devices.get(name=device)
    result = device_a.custom_fields["device_state"]

    return result


def get_states(devices):
    """Gets the state (device_state) stored in the Netbox for a list of devices."""

    result = {}
    for device in devices:
        device_a = utils.nb.dcim.devices.get(name=device)
        result[device] = device_a.custom_fields["device_state"]

    return result


def set_maintenance(device, state):
    """Set the maintenance state for a device in the Netbox."""

    logger.info(f"Set maintenance state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"maintenance": state}
    device_a.save()


def set_ironic_state(device, state):
    """Set the ironic state (ironic_state) for a device in the Netbox."""

    logger.info(f"Set ironic state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"ironic_state": state}
    device_a.save()


def set_introspection_state(device, state):
    """Set the introspection state (introspection_state) for a device in the Netbox."""

    logger.info(f"Set introspection state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"introspection_state": state}
    device_a.save()


def set_deployment_state(device, state):
    """Set the deployment state (deployment_state) for a device in the Netbox."""

    logger.info(f"Set deployment state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"deployment_state": state}
    device_a.save()


def set_device_state(device, state):
    """Set the state (device_state) for a device in the Netbox."""

    logger.info(f"Set state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"device_state": state}
    device_a.save()


def set_state(device, state, state_type):
    """Set the state for a device in the Netbox."""

    lock = Redlock(key=f"lock_state_{device}", masters={utils.redis})
    lock.acquire()

    if state_type == "power":
        set_power_state(device, state)
    elif state_type == "provision":
        set_provision_state(device, state)
    elif state_type == "introspection":
        set_introspection_state(device, state)
    elif state_type == "ironic":
        set_ironic_state(device, state)
    elif state_type == "deployment":
        set_deployment_state(device, state)
    else:
        set_device_state(device, state)

    lock.release()


def set_provision_state(device, state):
    """Set the provision state (provision_state) for a device in the Netbox."""

    logger.info(f"Set provision state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"provision_state": state}
    device_a.save()


def set_power_state(device, state):
    """Set the power state (power_state) for a device in the Netbox."""

    logger.info(f"Set power state of device {device} = {state}")

    device_a = utils.nb.dcim.devices.get(name=device)
    device_a.custom_fields = {"power_state": state}
    device_a.save()



# SPDX-License-Identifier: Apache-2.0

from cliff.command import Command
from loguru import logger
from redis import Redis

from osism import settings
from osism.tasks import conductor, netbox, reconciler, openstack, handle_task


redis = Redis(host=settings.REDIS_HOST, port=settings.REDIS_PORT, db=settings.REDIS_DB)
redis.ping()


class Ironic(Command):
    def get_parser(self, prog_name):
        parser = super(Ironic, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        # Get Ironic parameters from the conductor
        task = conductor.get_ironic_parameters.delay()
        task.wait(timeout=None, interval=0.5)
        ironic_parameters = task.get()

        # Add all unregistered systems from the Netbox in Ironic
        netbox.get_devices_not_yet_registered_in_ironic.apply_async(
            (), link=openstack.baremetal_create_nodes.s(ironic_parameters)
        )

        # Synchronize the current status in Ironic with the Netbox
        # openstack.baremetal_node_list.apply_async((), link=netbox.synchronize_device_state.s())

        # Remove systems from Ironic that are no longer present in the Netbox


class Sync(Command):
    def get_parser(self, prog_name):
        parser = super(Sync, self).get_parser(prog_name)
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the sync has been completed",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait

        task = reconciler.sync_inventory_with_netbox.delay()
        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)


class Manage(Command):
    def get_parser(self, prog_name):
        parser = super(Manage, self).get_parser(prog_name)
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the management of the netbox has been completed",
            action="store_true",
        )
        parser.add_argument(
            "--no-netbox-wait",
            default=False,
            help="Do not wait for the netbox API to be ready",
            action="store_true",
        )
        parser.add_argument(
            "--parallel",
            type=str,
            default=None,
            help="Process up to n files in parallel",
        )
        parser.add_argument(
            "--limit",
            type=str,
            default=None,
            help="Limit files by prefix",
        )
        parser.add_argument(
            "--skipdtl",
            default=False,
            help="Skip devicetype library",
            action="store_true",
        )
        parser.add_argument(
            "--skipmtl",
            default=False,
            help="Skip moduletype library",
            action="store_true",
        )
        parser.add_argument(
            "--skipres",
            default=False,
            help="Skip resources",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait
        arguments = []

        if parsed_args.no_netbox_wait:
            arguments.append("--no-wait")
        else:
            arguments.append("--wait")

        if parsed_args.parallel:
            arguments.append("--parallel")
            arguments.append(parsed_args.parallel)

        if parsed_args.limit:
            arguments.append("--limit")
            arguments.append(parsed_args.limit)

        if parsed_args.skipdtl:
            arguments.append("--skipdtl")
        else:
            arguments.append("--no-skipdtl")

        if parsed_args.skipmtl:
            arguments.append("--skipmtl")
        else:
            arguments.append("--no-skipmtl")

        if parsed_args.skipres:
            arguments.append("--skipres")
        else:
            arguments.append("--no-skipres")

        task_signature = netbox.manage.si(*arguments)
        task = task_signature.apply_async()
        if wait:
            logger.info(
                f"It takes a moment until task {task.task_id} (netbox-manager) has been started and output is visible here."
            )

        return handle_task(task, wait, format="script", timeout=3600)


class Ping(Command):
    def get_parser(self, prog_name):
        parser = super(Ping, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        task = netbox.ping.delay()
        task.wait(timeout=None, interval=0.5)
        result = task.get()
        print(result)
